//==============================================================================
// File: 			peentry.cpp
//
// Description: 	This utilty allows you to quickly locate the Virtual Address
//                of an entry point in a DLL or EXE file, or the PE's
//                execution entry point.
//
// Revisions: 		05/29/2000 - created
//
//==============================================================================
// Copyright(C) 2000, Tomas Restrepo. All rights reserved
// Send comments to: tomasr@mvps.org
//==============================================================================

#include "stdafx.h"
#include <dbghelp.h>
#pragma comment(lib, "dbghelp.lib")

const char * usage_string = "usage: peentry <options> <file>\noptions:\n"
                            "\t-e locate the PE's entry point\n"
                            "\t-f locate the function name\n"
                            "\t-i use case insensitive comparitions\n"
                            "\t-u undecorate exported function names";

struct options__ {
   std::string filename;
   std::string funcname;
   bool        find_entry;          // find PE's entry point
   bool        case_insensitive;    // case insensitive comparitions
   bool        sym_undecorate;      // undecorate symbol names
} options;


bool parse_args ( int argc, char * argv[] )
{
   bool err = false;
   options.filename   = "";
   options.funcname   = "";
   options.find_entry = false; 
   options.case_insensitive = false; 
   options.sym_undecorate = false;   

   for ( int i = 1; i < argc; i++ )
   {
      if ( argv[i][0] == '-' )
      {
         switch ( argv[i][1] )
         {
         case 'e':   options.find_entry = true;       break;
         case 'i':   options.case_insensitive = true; break;
         case 'u':   options.sym_undecorate = true;   break;
         case 'f':   
            options.find_entry = false;
            if ( i == argc-1 ) { err = true; break; }
            options.funcname = argv[++i];
            break;
         default:    err = true; break;
         }
      }
      else     // assume filename
         options.filename = argv[i];
   }
   if ( options.filename == "" ) err = true;
   return !err;
}

void get_entry_point ( void * base )
{
   IMAGE_NT_HEADERS * nthdr = NULL;
   ULONG_PTR entry = NULL;
   
   nthdr = ImageNtHeader ( base );
   entry = nthdr->OptionalHeader.ImageBase + 
           nthdr->OptionalHeader.AddressOfEntryPoint;

   printf ( "Entry point VA: 0x%p\n", entry );
}

void get_function ( void * base )
{
#define RvaToVa(x) ((ULONG_PTR)(nthdr->OptionalHeader.ImageBase+x))
#define RvaToEVa(x)((ULONG_PTR)((ULONG_PTR)(base)+x))

   IMAGE_NT_HEADERS *      nthdr   = NULL;
   IMAGE_DATA_DIRECTORY *  datadir = NULL;
   IMAGE_EXPORT_DIRECTORY * xprtdir;
   DWORD * eat;
   WORD * ordinaltable;
   ULONG_PTR * namestable;
   
   // find the data directory
   nthdr  = ImageNtHeader ( base );
   datadir = &nthdr->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT];
   
   // locate the exports directory, and it's tables....
   xprtdir = (IMAGE_EXPORT_DIRECTORY*)RvaToEVa(datadir->VirtualAddress);
   eat = (DWORD*)RvaToEVa(xprtdir->AddressOfFunctions);
   namestable = (ULONG_PTR*)RvaToEVa(xprtdir->AddressOfNames);
   ordinaltable = (WORD*)RvaToEVa(xprtdir->AddressOfNameOrdinals);
   
   //
   // now, find the function we are looking for
   // 
   
   // linear search
   const unsigned len = 1000;
   char name[len] = { 0x0 };

   DWORD n = xprtdir->NumberOfNames;
   for ( DWORD i = 0; i < n; i++ )
   {
      
      if ( !options.sym_undecorate || 
               !UnDecorateSymbolName ( 
                  (const char*)RvaToEVa(namestable[i]), 
                  name, len, UNDNAME_COMPLETE 
               )
         )
      {
         // use the decorated name
         strcpy ( name, (const char*)RvaToEVa(namestable[i]) );
      }
      bool equal = false;
      if ( options.case_insensitive )
         equal = (stricmp ( options.funcname.c_str ( ), name ) == 0);
      else
         equal = (strcmp  ( options.funcname.c_str ( ), name ) == 0);

      if ( equal )
      {
         WORD ord = ordinaltable[i];
         printf ( "Function's VA: 0x%p\n", RvaToVa(eat[ord]) ); 
         return;
      }
   }
   // didn't find it
   puts ( "Function name not found!" );
}

int main ( int argc, char * argv[] )
{
   if ( !parse_args ( argc, argv ) )
   {
      puts ( usage_string );
      return 1;
   }
   HMODULE base = NULL;
   base = LoadLibraryEx ( 
               options.filename.c_str ( ),
               NULL,
               DONT_RESOLVE_DLL_REFERENCES
            );
   if ( base != NULL )
   {
      if ( options.find_entry )
         get_entry_point ( base );
      else
         get_function ( base );
   }
   else
      printf ( "cannot open file - %lu\n", GetLastError ( ) );
   return 0;
}
